﻿
using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XELF {

	public class QuadFiller {
		readonly GraphicsDevice graphicsDevice;
		readonly VertexDeclaration declaration;
		readonly VertexBuffer vb;

		public QuadFiller(GraphicsDevice graphicsDevice) {
			this.graphicsDevice = graphicsDevice;
			vb = new VertexBuffer(graphicsDevice, typeof(VertexPositionTexture), 4, BufferUsage.WriteOnly);
			vb.SetData(new VertexPositionTexture[]{
					new VertexPositionTexture(new Vector3(-1, +1, 0), Vector2.Zero),
					new VertexPositionTexture(new Vector3(+1, +1, 0), Vector2.UnitX),
					new VertexPositionTexture(new Vector3(-1, -1, 0), Vector2.UnitY),
					new VertexPositionTexture(new Vector3(+1, -1, 0), Vector2.One),
				});
			declaration = new VertexDeclaration(graphicsDevice, VertexPositionTexture.VertexElements);
		}
		public void Fill(Effect effect) {
			graphicsDevice.RenderState.CullMode = CullMode.None;
			effect.Begin();
			foreach (var pass in effect.CurrentTechnique.Passes) {
				pass.Begin();
				Fill();
				pass.End();
			}
			effect.End();
		}
		public void Fill() {
			graphicsDevice.Vertices[0].SetSource(vb, 0, VertexPositionTexture.SizeInBytes);
			graphicsDevice.VertexDeclaration = declaration;
			graphicsDevice.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);
		}

	}

	public enum PostEffectMode {
		Hatch,
	}

	public class PostEffect : DrawableGameComponent {
		Effect effect;
		Random random = new Random();
		QuadFiller quad;

		PostEffectMode mode;

		public float EffectAngle = 0.5f;
		public float EffectAngleRange = 0.25f;
		public float EffectScale = 8;
		public Vector2 EffectTranslationRange = new Vector2(1, 1);

		public PostEffect(Game game)
			: base(game) {
		}

		public Texture2D SourceTexture {
			get { return effect.Parameters["SourceTexture"].GetValueTexture2D(); }
			set {
				effect.Parameters["SourceResolution"].SetValue(new Vector2(value.Width, value.Height));
				effect.Parameters["SourceTexture"].SetValue(value);
			}
		}

		void UpdateHatchTransform() {
			float x = (float)random.NextDouble() * EffectTranslationRange.X;
			float y = (float)random.NextDouble() * EffectTranslationRange.Y;
			float r = (float)((2 * Math.PI) * (random.NextDouble() * EffectAngleRange + EffectAngle - EffectAngleRange * 0.5));
			float c = (float)Math.Cos(r);
			float s = (float)Math.Sin(r);
			float a = GraphicsDevice.Viewport.AspectRatio;
			Matrix hatchTransform = new Matrix(
				c * EffectScale * a, -s * EffectScale * a, 0, 0,
				s * EffectScale, c * EffectScale, 0, 0,
				0, 0, 1, 0,
				x, y, 0, 1);
			effect.Parameters["HatchTransform"].SetValue(hatchTransform);
		}

		public Vector4 SourceColor {
			get { return effect.Parameters["SourceColor"].GetValueVector4(); }
			set { effect.Parameters["SourceColor"].SetValue(value); }
		}
		public Vector4 AmbientColor {
			get { return effect.Parameters["AmbientColor"].GetValueVector4(); }
			set { effect.Parameters["AmbientColor"].SetValue(value); }
		}
		public float EffectSensibility {
			get { return effect.Parameters["EffectSensibility"].GetValueSingle(); }
			set { effect.Parameters["EffectSensibility"].SetValue(value); }
		}
		public float EffectIntensity {
			get { return effect.Parameters["EffectIntensity"].GetValueSingle(); }
			set { effect.Parameters["EffectIntensity"].SetValue(value); }
		}

		protected override void LoadContent() {
			quad = new QuadFiller(GraphicsDevice);
			effect = Game.Content.Load<Effect>("Effects/Hatch");

			var hatchVolume = Game.Content.Load<Texture3D>("Textures/Hatch");
			effect.Parameters["HatchTexture"].SetValue(hatchVolume);
			effect.Parameters["HatchResolution"].SetValue(
				new Vector3(hatchVolume.Width, hatchVolume.Height, hatchVolume.Depth));
		}

		public PostEffectMode Mode {
			get { return mode; }
			set {
				mode = value;
				effect.CurrentTechnique = effect.Techniques[(int)mode];
			}
		}

		public override void Update(GameTime gameTime) {
			UpdateHatchTransform();
			base.Update(gameTime);
		}

		public void Draw(Texture2D source) {
			SourceTexture = source;
			quad.Fill(effect);
		}
	}

}
